/*
 * Configuration for Apollo provider, client, and caches.
 */

import { ApolloClient, HttpLink, from } from "@/apollo-boost";

import { RetryLink } from "@/apollo-link-retry";
import { persistCache, LocalStorageWrapper } from "@/apollo3-cache-persist";
import { InMemoryCache } from "@/apollo-cache-inmemory";

import gql from "@/graphql-tag";

import gqlSnackbarItems from "../components/app/snackbarItems.graphql";

// Cache for GraphQL query results in memory and persistent across sessions
const cache = new InMemoryCache();
await persistCache({
  cache: cache,
  storage: new LocalStorageWrapper(window.localStorage),
});

/**
 * Type definitions for Apollo/s local state management
 *  cf. https://www.apollographql.com/docs/react/local-state/local-state-management/
 */
const typeDefs = gql`
  type snackbarItem {
    id: ID!
    messageKey: String!
    color: String!
    read: Boolean!
  }

  type globalState {
    contentLoading: Boolean!
    browserTitle: String
  }

  type Mutation {
    toggleSnackbarItem(id: ID!): Boolean
  }
`;

/** Resolvers for local state management */
const resolvers = {
  Mutation: {
    /** Toggle snackbar item read or unread, given its ID */
    toggleSnackbarItem: (_, { id }, { cache }) => {
      const data = cache.readQuery({ query: gqlSnackbarItems });
      const currentItem = data.snackbarItems.find((item) => item.id === id);
      currentItem.read = !currentItem.read;
      cache.writeQuery({ query: gqlSnackbarItems, data });
      return currentItem.read;
    },
  },
};

/**
 * Utility function to add a snackbar on GraphQL errors.
 */
function addErrorSnackbarItem(messageKey) {
  let uuid = crypto.randomUUID();
  cache.writeQuery({
    query: gqlSnackbarItems,
    data: {
      snackbarItems: [
        {
          __typename: "snackbarItem",
          id: uuid,
          messageKey: messageKey,
          color: "red",
          read: false,
        },
      ],
    },
    variables: {
      id: uuid,
    },
  });
}

/**
 * Construct the GraphQL endpoint URI.
 *
 * @returns The URI of the GraphQL endpoint on the AlekSIS server
 */
function getGraphqlURL() {
  const settings = JSON.parse(
    document.getElementById("frontend_settings").textContent
  );
  const base = settings.urls.base || window.location.origin;
  return new URL(settings.urls.graphql, base);
}

// Define Apollo links for handling query operations.
const links = [
  // Automatically retry failed queries
  new RetryLink(),
  // Finally, the HTTP link to the real backend (Django)
  new HttpLink({
    uri: getGraphqlURL(),
  }),
];

/** Upstream Apollo GraphQL client */
const apolloClient = new ApolloClient({
  typeDefs,
  resolvers,
  cache,
  link: from(links),
});

const apolloOpts = {
  defaultClient: apolloClient,
  defaultOptions: {
    $query: {
      skip: (vm) => {
        // We only want to run this query when background activity is on and we are not reported offline
        return !vm.$root.backgroundActive || vm.$root.offline;
      },
      error: ({ graphQLErrors, networkError }, vm) => {
        if (graphQLErrors) {
          // Add a snackbar on all errors returned by the GraphQL endpoint
          console.error("A GraphQL query failed on the server");
          addErrorSnackbarItem("graphql.snackbar_error_message");
        }
        if (networkError) {
          // Set app offline globally on network errors
          //  This will cause the offline logic to kick in, starting a ping check or
          //  similar recovery strategies depending on the app/navigator state
          console.error(
            "Network error during GraphQL query, setting offline state"
          );
          vm.$root.offline = true;
        }
      },
    },
  },
};

export default apolloOpts;
